/*
 * kvadap_main.c -- kvadap char module
 *
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#define __NO_VERSION__ /* don't define kernel_verion in module.h */
#include <linux/module.h>
#include <linux/version.h>

char kernel_version [] = UTS_RELEASE;

#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        /* O_ACCMODE */

#include <asm/system.h>   /* cli(), *_flags */
#include <asm/segment.h>  /* memcpy and such */
#include <asm/io.h>

#include "kvadap.h"        /* local definitions */

/*********** local function prototypes ************/
int cvkey(Kvadap_Dev *dev,int k);

/************** variables **************/
int kvadap_major =   KVADAP_MAJOR;

Kvadap_Dev *kvadap_devices; /* allocated in init_module */
Kvadap_Prt_Dev *kvadap_prt_devices;
int kvadap_current_dev=0;

int kvadap_station=0;
int kvadap_state=0;
unsigned long poll_cnt=0;

/*
 * Open and close
 */

int kvadap_open (struct inode *inode, struct file *filp)
{
	filp->private_data = (void *)kvadap_current_dev;
    MOD_INC_USE_COUNT;
    return 0;          /* success */
}

void kvadap_release (struct inode *inode, struct file *filp)
{
    MOD_DEC_USE_COUNT;
}

/*
 * Data management: read and write
 */

read_write_t kvadap_read (struct inode *inode, struct file *filp,
                char *buf, count_t count)
{
    Kvadap_Dev *dev;
	int c;
	if ((int)filp->private_data >= 32)
		return -EINVAL;
	dev = &kvadap_devices[(int)filp->private_data];
	while (1) {
		while (dev->keyin == dev->keyout) {
			if (filp->f_flags & O_NONBLOCK)
				return -EAGAIN;
			interruptible_sleep_on(&dev->keyq);
			if (current->signal & ~current->blocked)
				return -ERESTARTSYS;
		}
/*		c = dev->keybuf[dev->keyout++];
*/
		c = cvkey(dev,(int)dev->keybuf[dev->keyout++]);

		if (dev->keyout >= KEY_MAX_CHARS) dev->keyout = 0;
		if (c) {
			put_user( c, buf++);
			break;
		}
	}	
    return 1;
}

read_write_t kvadap_write (struct inode *inode, struct file *filp,
                const char *buf, count_t count)
{
    Kvadap_Dev *dev;
    Kvadap_Prt_Dev *prtdev;
    int prttype = ((int)filp->private_data >= 32);
	int i=count;
	int tmp,k;
	prtdev = &kvadap_prt_devices[((int)filp->private_data) & 31];
	dev = &kvadap_devices[((int)filp->private_data) & 31];
	while(i--) {
		if (prttype) {
			tmp = prtdev->in+1;
			if (tmp>=PRT_MAX_CHARS) tmp=0;
			if (tmp == prtdev->out) {
				interruptible_sleep_on(&dev->prtq);
			}
			prtdev->buf[prtdev->in++] = get_user(buf++);
			if (prtdev->in >= PRT_MAX_CHARS) prtdev->in = 0;
		} else {
			tmp = dev->dspin+1;
			if (tmp>=DSP_MAX_CHARS) tmp=0;
			if (tmp == dev->dspout) {
				interruptible_sleep_on(&dev->dspq);
			}
			k = get_user(buf++);
			dev->dspbuf[dev->dspin++] = k;
			if (dev->dspin >= DSP_MAX_CHARS) dev->dspin = 0;
		}
	}
    return count;
}

int kvadap_select (struct inode *inode, struct file *filp,
	int mode, select_table *table)
{
    Kvadap_Dev *dev;
    Kvadap_Prt_Dev *prtdev;
    int prttype = ((int)filp->private_data >= 32);
	int tmp;
	prtdev = &kvadap_prt_devices[((int)filp->private_data) & 31];
	dev = &kvadap_devices[((int)filp->private_data) & 31];

	if (mode == SEL_OUT) {
		if (prttype) {
			tmp = prtdev->in+1;
			if (tmp>=PRT_MAX_CHARS) tmp=0;
 			if (tmp == prtdev->out) {
				select_wait( &dev->prtq, table );
				return 0;
			}
			return 1;
		} else {
			tmp = dev->dspin+1;
			if (tmp>=DSP_MAX_CHARS) tmp=0;
			if (tmp == dev->dspout) {
				select_wait( &dev->dspq, table );
				return 0;
			}
			return 1;
		}
	}
	if (mode == SEL_IN) {
		if (dev->keyin != dev->keyout)
			return 1;
		select_wait( &dev->keyq, table );
		return 0;
	}
	return 0;
}

int kvadap_ioctl (struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long arg)
{
    int err = 0, size = _IOC_SIZE(cmd); /* the size bitfield in cmd */

    /*
     * extract the type and number bitfields, and don't decode
     * wrong cmds: return EINVAL before verify_area()
     */
    if (_IOC_TYPE(cmd) != KVADAP_IOC_MAGIC) return -EINVAL;
    /*
     * the direction is a bitmask, and VERIFY_WRITE catches R/W
     * transfers. `Type' is user-oriented, while
     * verify_area is kernel-oriented, so the concept of "read" and
     * "write" is reversed
     */
    if (_IOC_DIR(cmd) & _IOC_READ)
        err = verify_area(VERIFY_WRITE, (void *)arg, size);
    else if (_IOC_DIR(cmd) & _IOC_WRITE)
        err =  verify_area(VERIFY_READ, (void *)arg, size);
    if (err) return err;

    switch(cmd) {

      case KVADAP_IOCRESET:
			kvadap_state = 0;
			outb_p(0x40,SERIAL_PORT);
			break;
      case KVADAP_IOCSADDR:
      		kvadap_current_dev = arg;
			filp->private_data = (void *)arg;
      case KVADAP_IOCGADDR:
				return (int)filp->private_data;
	  case KVADAP_IOCGPCNT:
	  			return poll_cnt;
      default:  /* redundant, as cmd was checked against MAXNR */
        return -EINVAL;
    }
    return 0;

}

#if LINUX_VERSION_CODE < VERSION_CODE(1,3,70)
void kvadap_interrupt(int irq, struct pt_regs *regs)
#else
void kvadap_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
	Kvadap_Dev *dev = &kvadap_devices[kvadap_station];
	Kvadap_Prt_Dev *prtdev = &kvadap_prt_devices[kvadap_station];
	int irqm,i;
	int k;
	irqm = inb_p(SERIAL_PORT+2);
	if ((irqm&7) == 2) { /* xmit holding reg empty */
		/****** transmitter empty interrupt ******/
		switch (kvadap_state) {
		case 1: /****** flushing display buffer ******/
			k = dev->dspbuf[dev->dspout++];
			if (dev->dspout >= DSP_MAX_CHARS)
				dev->dspout = 0;
			if (dev->dspout == dev->dspin) {
				kvadap_state = 3;
				wake_up_interruptible(&dev->dspq);
			}
			else 
				kvadap_state = 10;
			outb_p(k,SERIAL_PORT);
			break;
		case 10: /******* flushing to display prefix byte ******/
				 /* (select display data or command) */
			k = ((dev->dspbuf[dev->dspout] & 0x80) || 
				 (dev->dspbuf[dev->dspout] < 0x20) ? 0xc0 : 0xa0)+
				kvadap_station;
			outb_p(k,SERIAL_PORT);
			kvadap_state = 1;
			break;
		case 2: /****** flushing printer buffer ******/
			k = prtdev->buf[prtdev->out++];
			if (prtdev->out >= DSP_MAX_CHARS)
				prtdev->out = 0;
			dev->flags |= PRT_BUSY_FLAG;
			kvadap_state = 3;
			if (prtdev->out == prtdev->in)
				wake_up_interruptible(&dev->prtq);
			outb_p(k,SERIAL_PORT);
			break;			
		case 3: /****** send next polling ******/
			kvadap_state = 0;
			k = 0x40+kvadap_station;
			/*poll_cnt++;*/
			outb_p(k,SERIAL_PORT);
			break;
		case 0: /* just wait for receiver interrupt */
			break;
		}
	} else if ((irqm&7) == 4) { /* received data */
		/****** received data interrupt ******/
		k = inb_p(SERIAL_PORT);
		if (k==0xff) {
			if (dev->activity) dev->activity--;
		}
		else { 
			if (dev->activity < 100)
				dev->activity++;
			if (k==0xaa) {
				dev->flags &= ~PRT_BUSY_FLAG;
			} 
			else if (k) {
				poll_cnt++;
				dev->keybuf[dev->keyin++] = k;
				if (dev->keyin >= KEY_MAX_CHARS) 
					dev->keyin = 0;
				wake_up_interruptible(&dev->keyq);
			}
		}
		for (i=0;i<33;i++) {
			kvadap_station++;
			if (kvadap_station == 32)
				kvadap_station = 0;
			dev = &kvadap_devices[kvadap_station];
			prtdev = &kvadap_prt_devices[kvadap_station];
			if (dev->activity)
				break;
		}
		if (dev->dspout != dev->dspin) {
			k = ((dev->dspbuf[dev->dspout] & 0x80) || 
				 (dev->dspbuf[dev->dspout] < 0x20) ? 0xc0 : 0xa0)+
				kvadap_station;
			outb_p(k,SERIAL_PORT);
			kvadap_state = 1;
			return; 
		}
		if ((prtdev->out != prtdev->in) && 
			!(dev->flags & PRT_BUSY_FLAG)) {
				k = 0xe0+kvadap_station;
				outb_p(k,SERIAL_PORT);
				kvadap_state = 2;
				return;
		}
		k = 0x40+kvadap_station;
		kvadap_state = 0;
		outb_p(k,SERIAL_PORT);
	}
}

char scancode[] = "-----@--------`------q1---zsaw2--cxde43-- vftr5--nbhgy6---mju78--,kio09--./l;p----'-[=-----]-\\--------\x7f--1-47---0.2568\x1b--+3-*9--";
char scancode1[] = "-----R--------~------Q!---ZSAW@--CXDE$#-- VFTR%--NBHGY^---MJU&*--,KIO)(-->?L:P----\"-{+-----}-|----------------------------------";

int cvkey (Kvadap_Dev *dev,int k)
{
  int c;
  k &= 0xff;
  if (k == 0xe0) {
    dev->flags |= SPEC_FLAG;
    return 0;
  }
  if (k == 0xf0) {
    dev->flags |= REL_FLAG;
    return 0;
  }
  if (dev->flags & REL_FLAG) {
    if (k == 0x12)
      dev->flags &= ~MAI_FLAG;
    if (k == 0x14)
      dev->flags &= ~CTRL_FLAG;
	dev->flags &= ~(SPEC_FLAG | REL_FLAG); 
    return 0;
  }
  if (k == 0x12) {
	dev->flags |= MAI_FLAG;
    return 0;
  }
  if (k == 0x14) {
	dev->flags |= CTRL_FLAG;
    return 0;
  }
  k &= 0x7f;
  if (k == 0x5a)
    return 0x0a;
  if (dev->flags & SPEC_FLAG) {
    switch (k) {
    case 0x6b:
      return ((dev->flags & CTRL_FLAG) ? 0x81 : (0x80+'D'));
    case 0x74:
      return ((dev->flags & CTRL_FLAG) ? 0x82 : (0x80+'C'));
    case 0x75:
      return ((dev->flags & CTRL_FLAG) ? 0x83 : (0x80+'A'));
    case 0x72:
      return ((dev->flags & CTRL_FLAG) ? 0x84 : (0x80+'B'));
    case 0x70:
      return ((dev->flags & CTRL_FLAG) ? 0x85 : (0x80+'I'));
    case 0x6c:
      return ((dev->flags & CTRL_FLAG) ? 0x86 : (0x80+'H'));
    }
  }
  c = ((dev->flags & MAI_FLAG) ? scancode1[k] : scancode[k]);
  if (dev->flags & CTRL_FLAG)
    c &= 0x1f;
  return c;
}

struct file_operations kvadap_fops = {
    NULL,
    kvadap_read,
    kvadap_write,
    NULL,          /* kvadap_readdir */
    kvadap_select,		   
    kvadap_ioctl,
    NULL,          /* kvadap_mmap */
    kvadap_open,
    kvadap_release,
                   /* nothing more, fill with NULLs */
};
/*
 * Finally, the module stuff
 */

int init_module(void)
{
    int i,result;
	Kvadap_Dev *dev;

    result = register_chrdev(kvadap_major, "kvadap", &kvadap_fops);
    if (result < 0) {
        printk(KERN_WARNING "kvadap: can't get major %d\n",kvadap_major);
        return result;
    }
    if (kvadap_major == 0) kvadap_major = result; /* dynamic */

    kvadap_devices = kmalloc(32 * sizeof (Kvadap_Dev), GFP_KERNEL);
    if (!kvadap_devices) {
        result = -ENOMEM;
        goto fail_malloc;
    }
    memset(kvadap_devices, 0, 32 * sizeof (Kvadap_Dev));
	for (i=0;i<32;i++) {
		dev = &kvadap_devices[i];
		dev->activity = 3;
	} 
    kvadap_prt_devices = kmalloc(32 * sizeof (Kvadap_Prt_Dev), GFP_KERNEL);
    if (!kvadap_prt_devices) {
        result = -ENOMEM;
        goto fail_malloc1;
    }
    memset(kvadap_prt_devices, 0, 32 * sizeof (Kvadap_Prt_Dev));

#ifndef KVADAP_DEBUG
    register_symtab(NULL); /* otherwise, leave global symbols visible */
#endif

	/* initialize serial port */
	outb_p(0x87,SERIAL_PORT+3);
	outb_p(0x03,SERIAL_PORT); /* speed --> 38400 bps */
	outb_p(0x00,SERIAL_PORT+1);
	outb_p(0x07,SERIAL_PORT+3);
	outb_p(0x0b,SERIAL_PORT+4);

	/* setup interrupt service */
	if (request_irq(SERIAL_IRQ,kvadap_interrupt,
			SA_INTERRUPT,"kvadap",NULL)) {
		printk(KERN_INFO "kvadap: can't get assigned irq %i\n",SERIAL_IRQ);
		goto fail_irq;
	} else {
		outb_p(0x03,SERIAL_PORT+1); /* enable xmit & receive interrupts */
		inb_p(SERIAL_PORT);
		inb_p(SERIAL_PORT);
		inb_p(SERIAL_PORT);
		inb_p(SERIAL_PORT);
		inb_p(SERIAL_PORT+3);
		inb_p(SERIAL_PORT+4);
        inb_p(SERIAL_PORT+5);
        inb_p(SERIAL_PORT+6);
		outb_p(0x40,SERIAL_PORT); /* start polling */
	}
    return 0; /* succeed */
    
fail_irq:
	kfree(kvadap_prt_devices);
fail_malloc1:
  	kfree(kvadap_devices);
fail_malloc: 
	unregister_chrdev(kvadap_major, "kvadap");
    return result;
}

void cleanup_module(void)
{
	free_irq(SERIAL_IRQ,NULL);
    unregister_chrdev(kvadap_major, "kvadap");
    kfree(kvadap_devices);
    kfree(kvadap_prt_devices);
}
